@@ -32,7 +32,6 @@
 
 #include "nautilus-actions.h"
 #include "nautilus-application.h"
-#include "nautilus-bookmark-list.h"
 #include "nautilus-bookmarks-window.h"
 #include "nautilus-information-panel.h"
 #include "nautilus-main.h"
@@ -53,12 +52,13 @@
 #include <gtk/gtktogglebutton.h>
 #include <gtk/gtkvbox.h>
 #include <glib/gi18n.h>
-#include <libgnomevfs/gnome-vfs-uri.h>
+#ifdef HAVE_X11_XF86KEYSYM_H
+#include <X11/XF86keysym.h>
+#endif
 #include <libnautilus-private/nautilus-file-utilities.h>
 #include <libnautilus-private/nautilus-file-attributes.h>
 #include <libnautilus-private/nautilus-global-preferences.h>
 #include <libnautilus-private/nautilus-horizontal-splitter.h>
-#include <libnautilus-private/nautilus-icon-factory.h>
 #include <libnautilus-private/nautilus-metadata.h>
 #include <libnautilus-private/nautilus-mime-actions.h>
 #include <libnautilus-private/nautilus-program-choosing.h>
@@ -120,13 +120,26 @@ G_DEFINE_TYPE_WITH_CODE (NautilusWindow, nautilus_window, GTK_TYPE_WINDOW,
 			 G_IMPLEMENT_INTERFACE (NAUTILUS_TYPE_WINDOW_INFO,
 						nautilus_window_info_iface_init));
 
-static void
-icons_changed_callback (GObject *factory, NautilusWindow *window)
-{
-	g_return_if_fail (NAUTILUS_IS_WINDOW (window));
-
-	nautilus_window_update_icon (window);
-}
+static const struct {
+	unsigned int keyval;
+	const char *action;
+} extra_window_keybindings [] = {
+#ifdef HAVE_X11_XF86KEYSYM_H
+	{ XF86XK_AddFavorite,	NAUTILUS_ACTION_ADD_BOOKMARK },
+	{ XF86XK_Favorites,	NAUTILUS_ACTION_EDIT_BOOKMARKS },
+	{ XF86XK_Go,		NAUTILUS_ACTION_GO_TO_LOCATION },
+/* TODO?{ XF86XK_History,	NAUTILUS_ACTION_HISTORY }, */
+	{ XF86XK_HomePage,      NAUTILUS_ACTION_GO_HOME },
+	{ XF86XK_OpenURL,	NAUTILUS_ACTION_GO_TO_LOCATION },
+	{ XF86XK_Refresh,	NAUTILUS_ACTION_RELOAD },
+	{ XF86XK_Reload,	NAUTILUS_ACTION_RELOAD },
+	{ XF86XK_Search,	NAUTILUS_ACTION_SEARCH },
+	{ XF86XK_Start,		NAUTILUS_ACTION_GO_HOME },
+	{ XF86XK_Stop,		NAUTILUS_ACTION_STOP },
+	{ XF86XK_ZoomIn,	NAUTILUS_ACTION_ZOOM_IN },
+	{ XF86XK_ZoomOut,	NAUTILUS_ACTION_ZOOM_OUT }
+#endif
+};
 
 static void
 nautilus_window_init (NautilusWindow *window)
@@ -135,7 +148,7 @@ nautilus_window_init (NautilusWindow *window)
 	GtkWidget *menu;
 	GtkWidget *statusbar;
 
-	window->details = g_new0 (NautilusWindowDetails, 1);
+	window->details = G_TYPE_INSTANCE_GET_PRIVATE (window, NAUTILUS_TYPE_WINDOW, NautilusWindowDetails);
 
 	window->details->show_hidden_files_mode = NAUTILUS_WINDOW_SHOW_HIDDEN_FILES_DEFAULT;
 	
@@ -171,13 +184,6 @@ nautilus_window_init (NautilusWindow *window)
 			  GTK_EXPAND | GTK_FILL | GTK_SHRINK, 0,
 			  0,                                  0);
 
-	/* Register IconFactory callback to update the window border icon
-	 * when the icon-theme is changed.
-	 */
-	g_signal_connect_object (nautilus_icon_factory_get (), "icons_changed",
-				 G_CALLBACK (icons_changed_callback), window,
-				 0);
-
 	/* Register to menu provider extension signal managing menu updates */
 	g_signal_connect_object (nautilus_signaller_get_current (), "popup_menu_changed",
 			 G_CALLBACK (nautilus_window_load_extension_menus), window, G_CONNECT_SWAPPED);
@@ -224,20 +230,20 @@ nautilus_window_set_status (NautilusWindow *window, const char *text)
 }
 
 void
-nautilus_window_go_to (NautilusWindow *window, const char *uri)
+nautilus_window_go_to (NautilusWindow *window, GFile *location)
 {
 	g_return_if_fail (NAUTILUS_IS_WINDOW (window));
 
-	nautilus_window_open_location (window, uri, FALSE);
+	nautilus_window_open_location (window, location, FALSE);
 }
 
 
 void
-nautilus_window_go_to_with_selection (NautilusWindow *window, const char *uri, GList *new_selection)
+nautilus_window_go_to_with_selection (NautilusWindow *window, GFile *location, GList *new_selection)
 {
 	g_return_if_fail (NAUTILUS_IS_WINDOW (window));
 
-	nautilus_window_open_location_with_selection (window, uri, new_selection, FALSE);
+	nautilus_window_open_location_with_selection (window, location, new_selection, FALSE);
 }
 
 static gboolean
@@ -250,10 +256,8 @@ nautilus_window_go_up_signal (NautilusWindow *window, gboolean close_behind)
 void
 nautilus_window_go_up (NautilusWindow *window, gboolean close_behind)
 {
-	GnomeVFSURI *current_uri;
-	GnomeVFSURI *parent_uri;
+	GFile *parent;
 	GList *selection;
-	char *parent_uri_string;
 
 	g_return_if_fail (NAUTILUS_IS_WINDOW (window));
 
@@ -261,23 +265,19 @@ nautilus_window_go_up (NautilusWindow *window, gboolean close_behind)
 		return;
 	}
 	
-	current_uri = gnome_vfs_uri_new (window->details->location);
-	parent_uri = gnome_vfs_uri_get_parent (current_uri);
-	gnome_vfs_uri_unref (current_uri);
+	parent = g_file_get_parent (window->details->location);
 
-	if (parent_uri == NULL) {
+	if (parent == NULL) {
 		return;
 	}
 	
-	parent_uri_string = gnome_vfs_uri_to_string (parent_uri, GNOME_VFS_URI_HIDE_NONE);
-	gnome_vfs_uri_unref (parent_uri);
-
-	selection = g_list_prepend (NULL, g_strdup (window->details->location));
+	selection = g_list_prepend (NULL, g_object_ref (window->details->location));
 	
-	nautilus_window_open_location_with_selection (window, parent_uri_string, selection, close_behind);
+	nautilus_window_open_location_with_selection (window, parent, selection, close_behind);
 	
-	g_free (parent_uri_string);
-	eel_g_list_free_deep (selection);
+	g_object_unref (parent);
+	
+	eel_g_object_list_free (selection);
 }
 
 static void
@@ -357,13 +357,13 @@ nautilus_window_allow_reload (NautilusWindow *window, gboolean allow)
 void
 nautilus_window_go_home (NautilusWindow *window)
 {
-	char *home_uri;
+	GFile *home;
 
 	g_return_if_fail (NAUTILUS_IS_WINDOW (window));
 
-	home_uri = nautilus_get_home_directory_uri ();
-	nautilus_window_open_location (window, home_uri, FALSE);
-	g_free (home_uri);
+	home = g_file_new_for_path (g_get_home_dir ());
+	nautilus_window_open_location (window, home, FALSE);
+	g_object_unref (home);
 }
 
 void
@@ -377,11 +377,25 @@ nautilus_window_prompt_for_location (NautilusWindow *window,
 }
 
 char *
+nautilus_window_get_location_uri (NautilusWindow *window)
+{
+	g_return_val_if_fail (NAUTILUS_IS_WINDOW (window), NULL);
+
+	if (window->details->location) {
+		return g_file_get_uri (window->details->location);
+	}
+	return NULL;
+}
+
+GFile *
 nautilus_window_get_location (NautilusWindow *window)
 {
 	g_return_val_if_fail (NAUTILUS_IS_WINDOW (window), NULL);
 
-	return g_strdup (window->details->location);
+	if (window->details->location != NULL) {
+		return g_object_ref (window->details->location);
+	}
+	return NULL;
 }
 
 void
@@ -605,7 +619,9 @@ nautilus_window_finalize (GObject *object)
 
 	free_stored_viewers (window);
 
-	g_free (window->details->location);
+	if (window->details->location) {
+		g_object_ref (window->details->location);
+	}
 	eel_g_list_free_deep (window->details->pending_selection);
 
 	if (window->current_location_bookmark != NULL) {
@@ -622,9 +638,7 @@ nautilus_window_finalize (GObject *object)
 	}
 
 	g_free (window->details->title);
-	
-	g_free (window->details);
-	
+
 	G_OBJECT_CLASS (nautilus_window_parent_class)->finalize (object);
 }
 
@@ -725,6 +739,41 @@ nautilus_window_realize (GtkWidget *widget)
 	update_cursor (NAUTILUS_WINDOW (widget));
 }
 
+static gboolean
+nautilus_window_key_press_event (GtkWidget *widget,
+				 GdkEventKey *event)
+{
+	NautilusWindow *window;
+	int i;
+
+	window = NAUTILUS_WINDOW (widget);
+
+	for (i = 0; i < G_N_ELEMENTS (extra_window_keybindings); i++) {
+		if (extra_window_keybindings[i].keyval == event->keyval) {
+			const GList *action_groups;
+			GtkAction *action;
+
+			action = NULL;
+
+			action_groups = gtk_ui_manager_get_action_groups (window->details->ui_manager);
+			while (action_groups != NULL && action == NULL) {
+				action = gtk_action_group_get_action (action_groups->data, extra_window_keybindings[i].action);
+				action_groups = action_groups->next;
+			}
+
+			g_assert (action != NULL);
+			if (gtk_action_is_sensitive (action)) {
+				gtk_action_activate (action);
+				return TRUE;
+			}
+
+			break;
+		}
+	}
+
+	return GTK_WIDGET_CLASS (nautilus_window_parent_class)->key_press_event (widget, event);
+}
+
 /*
  * Main API
  */
@@ -737,7 +786,8 @@ free_activate_view_data (gpointer data)
 	activate_data = data;
 
 	g_free (activate_data->id);
-	g_free (activate_data);
+
+	g_slice_free (ActivateViewData, activate_data);
 }
 
 static void
@@ -759,16 +809,16 @@ add_view_as_menu_item (NautilusWindow *window,
 {
 	const NautilusViewInfo *info;
 	GtkRadioAction *action;
-	char *action_name;
+	char action_name[32];
 	ActivateViewData *data;
 
-	char *accel;
-	char *accel_path;
+	char accel[32];
+	char accel_path[48];
 	unsigned int accel_keyval;
 
 	info = nautilus_view_factory_lookup (identifier);
 	
-	action_name = g_strdup_printf ("view_as_%d", index);
+	g_snprintf (action_name, sizeof (action_name), "view_as_%d", index);
 	action = gtk_radio_action_new (action_name,
 				       _(info->view_as_label_with_mnemonic),
 				       _(info->display_location_label),
@@ -776,18 +826,14 @@ add_view_as_menu_item (NautilusWindow *window,
 				       0);
 
 	if (index >= 1 && index <= 9) {
-		accel = g_strdup_printf ("%d", index);
-
-		accel_path = g_strdup_printf ("<Nautilus-Window>/%s", action_name);
+		g_snprintf (accel, sizeof (accel), "%d", index);
+		g_snprintf (accel_path, sizeof (accel_path), "<Nautilus-Window>/%s", action_name);
 
 		accel_keyval = gdk_keyval_from_name (accel);
 		g_assert (accel_keyval != GDK_VoidSymbol);
 
 		gtk_accel_map_add_entry (accel_path, accel_keyval, GDK_CONTROL_MASK);
 		gtk_action_set_accel_path (GTK_ACTION (action), accel_path);
-
-		g_free (accel);
-		g_free (accel_path);
 	}
 
 	if (window->details->view_as_radio_action != NULL) {
@@ -799,7 +845,7 @@ add_view_as_menu_item (NautilusWindow *window,
 		window->details->view_as_radio_action = action;
 	}
 
-	data = g_new (ActivateViewData, 1);
+	data = g_slice_new (ActivateViewData);
 	data->window = window;
 	data->id = g_strdup (identifier);
 	g_signal_connect_data (action, "activate",
@@ -817,8 +863,7 @@ add_view_as_menu_item (NautilusWindow *window,
 			       action_name,
 			       GTK_UI_MANAGER_MENUITEM,
 			       FALSE);
-	g_free (action_name);
-	
+
 	return action; /* return value owned by group */
 }
 
@@ -898,7 +943,7 @@ static void
 nautilus_window_synch_view_as_menus (NautilusWindow *window)
 {
 	int index;
-	char *action_name;
+	char action_name[32];
 	GList *node;
 	GtkAction *action;
 
@@ -921,10 +966,9 @@ nautilus_window_synch_view_as_menus (NautilusWindow *window)
 		remove_extra_viewer_in_view_as_menus (window);
 	}
 
-	action_name = g_strdup_printf ("view_as_%d", index);
+	g_snprintf (action_name, sizeof (action_name), "view_as_%d", index);
 	action = gtk_action_group_get_action (window->details->view_as_action_group,
 					      action_name);
-	g_free (action_name);
 
 	/* Don't trigger the action callback when we're synchronizing */
 	g_signal_handlers_block_matched (action,
@@ -972,22 +1016,18 @@ real_load_view_as_menu (NautilusWindow *window)
 					  window->details->short_list_merge_id);
 		window->details->short_list_merge_id = 0;
 	}
+	if (window->details->extra_viewer_merge_id != 0) {
+		gtk_ui_manager_remove_ui (window->details->ui_manager,
+					  window->details->extra_viewer_merge_id);
+		window->details->extra_viewer_merge_id = 0;
+		window->details->extra_viewer_radio_action = NULL;
+	}
 	if (window->details->view_as_action_group != NULL) {
 		gtk_ui_manager_remove_action_group (window->details->ui_manager,
 						    window->details->view_as_action_group);
 		window->details->view_as_action_group = NULL;
 	}
 
-	if (window->details->extra_viewer_merge_id != 0) {
-		gtk_ui_manager_remove_ui (window->details->ui_manager,
-					  window->details->extra_viewer_merge_id);
-		window->details->extra_viewer_merge_id = 0;
-	}
-	if (window->details->extra_viewer_radio_action != NULL) {
-		gtk_action_group_remove_action (window->details->view_as_action_group,
-						GTK_ACTION (window->details->extra_viewer_radio_action));
-		window->details->extra_viewer_radio_action = NULL;
-	}
 	g_free (window->details->extra_viewer);
 	window->details->extra_viewer = NULL;
 	
@@ -1046,7 +1086,7 @@ nautilus_window_load_view_as_menus (NautilusWindow *window)
 
         g_return_if_fail (NAUTILUS_IS_WINDOW (window));
 
-	attributes = nautilus_mime_actions_get_full_file_attributes ();
+	attributes = nautilus_mime_actions_get_required_file_attributes ();
 
 	cancel_view_as_callback (window);
 	nautilus_file_call_when_ready (window->details->viewed_file,
@@ -1081,7 +1121,7 @@ real_get_title (NautilusWindow *window)
         }
         
 	if (title == NULL) {
-                title = nautilus_compute_title_for_uri (window->details->location);
+                title = nautilus_compute_title_for_location (window->details->location);
         }
 
 	return title;
@@ -1177,16 +1217,30 @@ nautilus_window_update_title (NautilusWindow *window)
 void
 nautilus_window_update_icon (NautilusWindow *window)
 {
-	char *icon_name;
+	NautilusIconInfo *info;
+	const char *icon_name;
+	GdkPixbuf *pixbuf;
 
 	g_return_if_fail (NAUTILUS_IS_WINDOW (window));
 
-	icon_name = EEL_CALL_METHOD_WITH_RETURN_VALUE (NAUTILUS_WINDOW_CLASS, window,
-						       get_icon_name, (window));
+	info = EEL_CALL_METHOD_WITH_RETURN_VALUE (NAUTILUS_WINDOW_CLASS, window,
+						 get_icon, (window));
 
-	if (icon_name != NULL) {
-		gtk_window_set_icon_name (GTK_WINDOW (window), icon_name);
-		g_free (icon_name);
+	icon_name = NULL;
+	if (info) {
+		icon_name = nautilus_icon_info_get_used_name (info);
+		if (icon_name != NULL) {
+			gtk_window_set_icon_name (GTK_WINDOW (window), icon_name);
+		} else {
+			pixbuf = nautilus_icon_info_get_pixbuf_nodefault (info);
+			
+			if (pixbuf) {
+				gtk_window_set_icon (GTK_WINDOW (window), pixbuf);
+				g_object_unref (pixbuf);
+			} 
+		}
+		
+		g_object_unref (info);
 	}
 }
 
@@ -1292,7 +1346,9 @@ nautilus_window_set_viewed_file (NautilusWindow *window,
 	}
 
 	if (file != NULL) {
-		attributes = NAUTILUS_FILE_ATTRIBUTE_DISPLAY_NAME | NAUTILUS_FILE_ATTRIBUTE_SLOW_MIME_TYPE;
+		attributes =
+			NAUTILUS_FILE_ATTRIBUTE_INFO |
+			NAUTILUS_FILE_ATTRIBUTE_LINK_INFO;
 		nautilus_file_monitor_add (file, window, attributes);
 	}
 
@@ -1383,25 +1439,25 @@ add_to_history_list (NautilusBookmark *bookmark)
 }
 
 void
-nautilus_remove_from_history_list_no_notify (const char *uri)
+nautilus_remove_from_history_list_no_notify (GFile *location)
 {
 	NautilusBookmark *bookmark;
 
-	bookmark = nautilus_bookmark_new (uri, "");
+	bookmark = nautilus_bookmark_new (location, "");
 	remove_from_history_list (bookmark);
 	g_object_unref (bookmark);
 }
 
 gboolean
-nautilus_add_to_history_list_no_notify (const char *uri,
+nautilus_add_to_history_list_no_notify (GFile *location,
 					const char *name,
 					gboolean has_custom_name,
-					const char *icon)
+					GIcon *icon)
 {
 	NautilusBookmark *bookmark;
 	gboolean ret;
 
-	bookmark = nautilus_bookmark_new_with_icon (uri, name, has_custom_name, icon);
+	bookmark = nautilus_bookmark_new_with_icon (location, name, has_custom_name, icon);
 	ret = add_to_history_list (bookmark);
 	g_object_unref (bookmark);
 
@@ -1545,7 +1601,7 @@ nautilus_window_info_iface_init (NautilusWindowInfoIface *iface)
 	iface->get_title = nautilus_window_get_cached_title;
 	iface->get_history = nautilus_window_get_history;
 	iface->get_bookmark_list = nautilus_window_get_bookmark_list;
-	iface->get_current_location = nautilus_window_get_location;
+	iface->get_current_location = nautilus_window_get_location_uri;
 	iface->get_ui_manager = nautilus_window_get_ui_manager;
 	iface->get_selection_count = nautilus_window_get_selection_count;
 	iface->get_selection = nautilus_window_get_selection;
@@ -1566,6 +1622,7 @@ nautilus_window_class_init (NautilusWindowClass *class)
 	GTK_WIDGET_CLASS (class)->show = nautilus_window_show;
 	GTK_WIDGET_CLASS (class)->size_request = nautilus_window_size_request;
 	GTK_WIDGET_CLASS (class)->realize = nautilus_window_realize;
+	GTK_WIDGET_CLASS (class)->key_press_event = nautilus_window_key_press_event;
 	class->add_current_location_to_history_list = real_add_current_location_to_history_list;
 	class->get_title = real_get_title;
 	class->set_title = real_set_title;
@@ -1628,6 +1685,8 @@ nautilus_window_class_init (NautilusWindowClass *class)
 			     "\n"
 			     "    widget \"*.nautilus-extra-view-widget\" style:rc \"nautilus-extra-view-widgets-style-internal\" \n"
 			     "\n");
+
+	g_type_class_add_private (G_OBJECT_CLASS (class), sizeof (NautilusWindowDetails));
 }
 
 /**
